/*  plumb - the pipe shell - virtual file descriptors
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "plumb.h"
#include "vp.h"
#include "rp.h"
#include "fd.h"
#include "config.h"

fd_t *fd_ref_(proc_virt_t *proc, int fd_num)
{
	fd_t *fd;

	if (fd_num < 0)
		fd_num = proc->h.num_fds;

	if (fd_num >= proc->h.alloced_fds) {
		int n;
		fd_t *new;
//		reCalloc(proc->h.fds, sizeof(fd_t), proc->h.alloced_fds, (fd_num + 16));
		new = calloc(sizeof(fd_t), (fd_num + 16));
		memcpy(new, proc->h.fds, proc->h.alloced_fds * sizeof(fd_t));
		for(n = 0; n < proc->h.alloced_fds; n++) {
			pipe_t *p = proc->h.fds[n].pipe;
			if (p != NULL) {
				if (p->from == &(proc->h.fds[n]))
					p->from = &(new[n]);
				else if (p->to == &(proc->h.fds[n]))
					p->to = &(new[n]);
				else
					plumb_error("alien fd in fd_ref_");
			}
		}
		free(proc->h.fds);
		proc->h.fds = new;
		
		if (!proc->h.virt) {
			proc_real_t *pr = (proc_real_t *)proc;
			reCalloc(pr->sock, sizeof(P_net_socket), pr->h.alloced_fds, (fd_num + 16));
		}
		proc->h.alloced_fds = fd_num + 16;
	}
	if (fd_num >= proc->h.num_fds)
		proc->h.num_fds = fd_num + 1;

	fd = &proc->h.fds[fd_num];
	if (fd->proc.v == NULL) {
		fd->proc.v = proc;
		fd->fd     = fd_num;
		fd->pi     = -1;
	}
	else
		assert(fd->fd == fd_num);
	return fd;
}

/* free buffers of a closed fd and clear buffering-related states */
static void fd_clear_buff(fd_t *fd)
{
	if (fd->obuff != NULL) {
		free(fd->obuff);
		fd->obuff = NULL;
		fd->obuff_len = 0;
	}
	if (fd->rb != NULL) {
		read_buf_t *r;
		do {
			r = fd->rb;
			fd->rb = r->next;
			free(r);
		} while(fd->rb != NULL);
	}
	fd->close_after_write = 0;
}

static void fd_close_event(char *procid, int fd)
{
	plumb_event(strlen(procid) + 128, "plumb_close %s:%d\n", procid, fd);
}

void fd_close(fd_t *fd)
{
	pipe_t *p = fd->pipe;
	proc_virt_t *pv;
	proc_real_t *pr;
	P_net_socket socket;

	if ((p == NULL) || (fd->closing))
		return;

	fd_close_event(fd->proc.v->h.id, fd->fd);

	fd->closing = 1;
	if (p->from == fd) {            /* source got closed, notify sink */
		if (p->to != NULL) {
		pv = p->to->proc.v;
		pr = p->to->proc.r;
		fd_close_event(pr->h.id, p->to->fd);
		if (!pv->h.virt) {              /* sink is a real process: shut down and disable watch (that might have been set to watch write) */
			if (p->to->obuff == NULL) {
				p->to->closing=1;
				P_net_shutdown(pr->sock[p->to->fd], P_SHUT_WR);
				fd_del_watch(p->to, P_POLLOUT);
				
			}
			else {
				p->to->close_after_write = 1;
			}
		}
		else {                         /* sink is a virtual process: notify */
			p->to->closing=1;
			if ((pv->sink_eof != NULL) && (!p->to->eof))
				vp_act(pv, pv->sink_eof(pv, p->to->fd));
		}
		/* close this end, if it's a real socket */
		socket = fd_query_socket(fd);
		if (socket != P_NET_INVALID_SOCKET) {
			rp_socket_inval(fd->proc.r, socket);
			P_net_shutdown(socket, P_SHUT_RD);
			fd_clear_buff(fd);
			fd->to_delete = 1;
		}
		}
	}
	else if (p->to == fd) {         /* sink got closed, back promote */
		if (p->from != NULL) {
		pv = p->from->proc.v;
		pr = p->from->proc.r;
		fd_close_event(pr->h.id, p->from->fd);
		p->from->closing = 1;
		if (!pv->h.virt) {              /* source is a real process: shut down and disable watch (that might have been set to watch read) */
			P_net_close(pr->sock[p->from->fd]);
			p->from->to_delete = 1;
			fd_del_watch(p->from, P_POLLIN);
		}
		else                          /* sink is a virtual process: notify */
			if ((pv->source_eof != NULL) && (!p->to->eof))
				vp_act(pv, pv->source_eof(pv, p->to->fd));

		/* close this end, if it's a real socket */
		socket = fd_query_socket(fd);
		if (socket != P_NET_INVALID_SOCKET) {
			rp_socket_inval(fd->proc.r, socket);
			P_net_shutdown(socket, P_SHUT_WR);
			fd_del_watch(fd, P_POLLOUT);
			fd_clear_buff(fd);
			fd->to_delete = 1;
		}
		}
	}
	else
		plumb_error("alien fd for pipe in fd_eof()\n");

	if (p->from != NULL)
		p->from->eof = 1;

	if (p->to != NULL)
		p->to->eof = 1;

	if ((p->from != NULL) && (p->from->proc.v != NULL))
		proc_kill_on_eof(p->from->proc.v);
	if ((p->to != NULL) && (p->to->proc.v != NULL))
		proc_kill_on_eof(p->to->proc.v);
}

int fd_write(fd_t *fd, const char *data, int len)
{
	pipe_t *p = fd->pipe;
	proc_virt_t *pv;
	proc_real_t *pr;

	/* if fd is not connected at all, write fails */
	if (p == NULL)
		return -1;

	if (fd->eof)
		return -1;

	if (p->from == fd)            /* read event at source */
		fd = p->to;
	else if (p->to != fd)         /* or direct write on the sink? */
		plumb_error("alien fd for pipe in fd_write()\n");

	/* fd is where we need to write */
	pv = fd->proc.v;
	pr = fd->proc.r;
	if (pv->h.virt) {
		if (pv->input != NULL)
			vp_act(pv, pv->input(pv, fd->fd, data, len));
	}
	else {
		int written;
		if (fd->obuff == NULL)
			written = P_net_write(pr->sock[fd->fd], (void *)data, len);
		else {
			written = 0;
/*			fprintf(stderr, "APPEND!\n");*/
		}
#ifdef RW_DEBUG
		{
			char *s;
			s = malloc(len+1);
			memcpy(s, data, len);
			s[len] = '\0';
			fprintf(stderr, "write: [%d]: %d -> %d %s: ", pr->sock[fd->fd], len, written, fd->obuff == NULL ? "real" : "buff");
			fprintstr(stderr, data);
			fprintf(stderr, "\n");
			free(s);
		}
#endif
		if ((written < 0) && (!P_net_inprogress)) {
			/* write error is an eof, except the event generated */
			plumb_event(strlen(pr->h.id) + 128, "plumb_write_error %s:%d\n", pr->h.id, fd->fd);
			fd_close(fd);
		}
		else if (written < len) {
			int left;
			/* short write - buffer the rest */
			if (written < 0)
				written  = 0;
			left = len - written;
/*			we used to have an error for if (fd->obuff != NULL)  here, but...
				 writing while obuff is not NULL happens in a "RP | VP | RP" chain if
				 VP adds considerable amount of data. In this case we need to store
				 the data received from the VP somewhere. The easiest way is to store
				 it here and do the usual back_blocking. */
			fd->obuff = realloc(fd->obuff, fd->obuff_len + left);
			memcpy(fd->obuff + fd->obuff_len, data+written, left);
			fd->obuff_len += left;
			fd_add_watch(fd, P_POLLOUT);
			back_block(fd, 1);
		}
		else {
			/* succesful write */
			if (fd->close_after_write) {
				fd->close_after_write = 0;
				fd_close(fd);
			}
		}
	}
	return 0;
}

void fd_rp_source_poll(fd_t *src)
{
	if ((src->paused) || (src->blocked) || (src->eof)) {
		fd_del_watch(src, P_POLLIN);
	}
	else
		fd_add_watch(src, P_POLLIN);
}

static void fd_vp_source_poll(fd_t *src, fd_t *sink)
{
	int is_blocking;
	proc_virt_t *sproc = src->proc.v;

	is_blocking = ((src->blocked) || (sink->paused) || (sink->eof));
	if (src->vp_last_blstate != is_blocking) {
		src->vp_last_blstate = is_blocking;
		vp_act(sproc, sproc->source_block(sproc, src->fd, is_blocking));
	}
}

void fd_flush_rb(fd_t *src)
{
				read_buf_t *r;
				/* if rb is not NULL, the fd is already HUP'd, there's no reason to
				   restore POLLIN; instead, try flushing pending input until output
				   blocks again. This can be detected only by setting POLLIN and
				   looking at it... */
				fd_add_watch(src, P_POLLIN); /* for watching block */

				/* write records as long as we are not blocked and did not run out of records */
				do {
					r = src->rb;
					fd_write(src, r->data, r->len);
					src->rb = r->next;
					free(r);
				} while(fd_query_watch(src, P_POLLIN) && (src->rb != NULL));

				/* disable pollin as we are blocked again, or going to close */
				src->blocked = 1;

				/* if we don't have anything else to send from the input buffer, close socket */
				if (src->rb == NULL)
					fd_close(src);
}


void back_block(fd_t *sink, int is_blocking)
{
	fd_t *src = sink->pipe->from;
	proc_virt_t *sproc = src->proc.v;
	assert(sink->pipe->to == sink);
	assert(sink->pipe->from != sink);

	src->blocked = is_blocking;

	/* block release while in pause - do not backpromote */
	if ((!is_blocking) && (src->paused))
		return;

	if (sproc->h.virt) {
		/* for a virtual, just call the handler */
		if ((sproc->source_block != NULL) && (!src->closing))
			fd_vp_source_poll(src, sink);
	}
	else {
		/* for a real, stop or start reading */
		if (!is_blocking) {
			/* real fd unblocked */
			if (src->rb != NULL)
				fd_flush_rb(src);
			else
				src->blocked = 0;
		}
		fd_rp_source_poll(src);
	}
}

/* Pause: do not copy the flag, mark fds at the paused process, but block
   sources for paused sinks */
void fd_pause(fd_t *fd, int is_paused)
{
	pipe_t *p;

	/* no change in state */
	if (fd->paused == is_paused)
		return;

	fd->paused = is_paused;
	p = fd->pipe;
	if ((p != NULL) && (p->to == fd)) {
		/* fd is a sink - if paused, back-block; if unpaused and not blocking, back-unblock */
		if (is_paused)
			back_block(fd, 1);
		else if (!fd->blocked)
			back_block(fd, 0);
	}

	if ((p != NULL) && (p->from == fd)) {
		/* source gets paused */
		proc_virt_t *vp = fd->proc.v;
		if (!vp->h.virt)
			fd_rp_source_poll(fd);

	/* if an vp pipe gets paused, ignore it - when pausing a vp, it will block
	its _sources_ and any rp or vp feeding the sources, so no data will
	be processed. On top of this, VPs should not emit on paused sources */
	}

	/* if an unbound pipe gets paused, same */
}

