/*  plumb - the pipe shell - portable 'busybox' for regression tests
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "libporty/util/buf.h"
#include "libporty/util/argv.h"
#include "libporty/host/time.h"

void cmd_sleep(char *dur)
{
	double d;
	char *end;

	d = strtod(dur, &end);
	if (*end != '\0') {
		fprintf(stderr, "plumbtest: invalid duration for sleep - must be a number\n");
		exit(1);
	}
	P_usleep(d * 1000000.0);
}

void cmd_exit(char *code)
{
	exit(atoi(code));
}

void cmd_echo(char *text)
{
	puts(text);
	fflush(stdout);
}

void cmd_seq(char *cnt, char *msg)
{
	int count, num;
	char *end;

	count = strtol(cnt, &end, 10);
	if (*end != '\0') {
		fprintf(stderr, "plumbtest: invalid count for seq - must be a number\n");
		exit(1);
	}

	num = 1;
	if (msg != NULL) {
		while(count > 0) {
			printf("%d %s\n", num++, msg);
			count--;
		}
	}
	else {
		while(count > 0) {
			printf("%d\n", num++);
			count--;
		}
	}
}

void cmd_cat(char *fn, char *prefix, int sync)
{
	FILE *f;
	int nl;

	if ((fn != NULL) && (*fn != '\0') && (*fn != '-')) {
		f = fopen(fn, "r");
		if (f == NULL) {
			fprintf(stderr, "plumbtest: can't open %s for reading\n", fn);
			exit(1);
		}
	}
	else
		f = stdin;

	nl = 1;
	/* minimal buffering is reading byte-by-byte */
	for(;;) {
		int c;
		c = fgetc(f);
		if (c < 0)
			return;
		if ((c == '\r') || (c == '\n'))
			nl = 1;
		else if ((nl) && (prefix != NULL)) {
			printf("%s", prefix);
			nl = 0;
		}
		fputc(c, stdout);
		if (sync)
			fflush(stdout);
	}
}

void cmd_alnumcount()
{
	int count = 0;
	for(;;) {
		int c;
		c = fgetc(stdin);
		if (c < 0) {
			printf("%d\n", count);
			return;
		}
		if (isalnum(c))
			count++;
	}
}

void cmd_chrcount(char chr)
{
	int count = 0;
	for(;;) {
		int c;
		c = fgetc(stdin);
		if (c < 0) {
			printf("%d\n", count);
			return;
		}
		if (c == chr)
			count++;
	}
}

#define rest(from) (argv[from] == NULL ? NULL : (statement + (argv[from] - argv[0])))

void execute(char *statement, int argc, char *argv[])
{
	if (strcmp(argv[0], "sleep") == 0)           cmd_sleep(argv[1]);
	else if (strcmp(argv[0], "echo") == 0)       cmd_echo(rest(1));
	else if (strcmp(argv[0], "exit") == 0)       cmd_exit(argv[1]);
	else if (strcmp(argv[0], "seq") == 0)        cmd_seq(argv[1], argv[2]);
	else if (strcmp(argv[0], "cat") == 0)        cmd_cat(argv[1], rest(2), 1);
	else if (strcmp(argv[0], "catbuff") == 0)    cmd_cat(argv[1], rest(2), 0);
	else if (strcmp(argv[0], "alnumcount") == 0) cmd_alnumcount();
	else if (strcmp(argv[0], "linecount") == 0)  cmd_chrcount('\n');
	else {
		printf("plumbtest: unknown command %s\n", argv[0]); /* use stdout because plumb may eat up stderr */
		exit(1);
	}
}

char *strclone(const char *s)
{
	char *out;
	int len;

	assert(s != NULL);

	len = strlen(s);
	out = malloc(len+1);
	memcpy(out, s, len+1);
	return out;
}

int main(int argc, char *argv[])
{
	int n;
	P_buf_t *b;
	char *statement;

	b = P_buf_init("&");

	for(n = 1; n < argc; n++) {
		P_buf_fill(b, argv[n]);
		if (n < argc - 1)
			P_buf_fill(b, " ");
	}

	P_buf_fill(b, "&");

	while((statement = P_buf_get(b)) != NULL) {
		char *args, *stm;
		char *argv[256];
		int argc;
		
		memset(argv, 0, sizeof(argv));
		stm = statement;
		while((*stm == ' ') || (*stm == '\t')) stm++;
		args = strclone(stm);
		P_argv_split(args, &argc, argv, sizeof(argv) / sizeof(char *), " \t");
		execute(stm, argc, argv);

		free(args);
		P_buf_free(statement);
	}
	return 0;
}
