/*  plumb - the pipe shell - common process helpers for VPs and RPs
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "genht/htsp.h"
#include "genht/hash.h"
#include "process.h"
#include "parse_common.h"
#include "plumb.h"
#include "vp.h"
#include "rp.h"
#include "fd.h"
#include "pipe.h"

static void proc_prop_set(proc_virt_t *proc, kvlist_t *props);

/* a process is either virtual or real - this structure is used for the internal hash */
typedef union proc_u {
	void *ptr;
	proc_virt_t *v;
	proc_real_t *r;
} proc_t;

htsp_t *procs = NULL;

static void procs_init(void)
{
	if (procs == NULL)
		procs = htsp_alloc(strhash, strkeyeq);
}

/* create a new process, return fd for stdout */
proc_virt_t *proc_create(transaction_t *tx, char *id, int virt, char *cmd, kvlist_t *args, kvlist_t *props)
{
	proc_t p;

	/* id use: vp_alloc() and rp_alloc() will strclone() it, original goes in the hash so
	   must be strclone'd before the call */

	procs_init();

	if (htsp_get(procs, (char *)id) != NULL)
		plumb_error("multiple definition of process %s\n", id);

	if (virt) {
		p.v = vp_alloc(tx, cmd, id, args);
		if (p.v == NULL)
			plumb_error("can't create virtual process %s: unknown command %s\n", id, cmd);
	} else {
		p.r = rp_alloc(tx, cmd, id, args);
		if (p.r == NULL)
			plumb_error("can't create real process %s (internal error)\n", id);
	}

	/* properties are already set to their defaults by vp_alloc() and rp_alloc()
	   (because defaults are different per process class). Time to overwrite
	   specified properties from the list */
	proc_prop_set(p.v, props);
	p.v->h.state = PS_STOPPED;
	htsp_set(procs, id, p.ptr);

	return p.v;
}

void proc_free(proc_virt_t *p)
{
	htsp_entry_t *e;
	int n;

	e = htsp_popentry(procs, p->h.id);

	for(n = 0; n < p->h.num_fds; n++) {
		pipe_t *pipe = p->h.fds[n].pipe;
		if (pipe != NULL) {
			int m;
			/* the same pipe may be referenced multiple times: a loop from
			   one fd to another of the same process - make sure such pipes
			   are not double freed */
			for(m = n+1; m < p->h.num_fds; m++) {
				if (p->h.fds[m].pipe == pipe)
					p->h.fds[m].pipe = NULL;
			}
			fd_close(&p->h.fds[n]);
			pipe_unbind(pipe, p);
		}
	}

	free(e->key);

	if (p->h.tx_name != NULL);
		free(p->h.tx_name);
	free(p->h.id);

	if (p->h.fds != NULL)
		free(p->h.fds);

	/* free type specific fields */
	if (p->h.virt)
		vp_free(p);
	else
		rp_free((proc_real_t *)p);
	free(p);
}

int proc_lookup(const char *id, proc_virt_t **pv, proc_real_t **pr)
{
	proc_t p;

	/* make sure output pointers are NULL for the item(s) we won't find */
	if (pr != NULL)
		*pr = NULL;
	if (pv != NULL)
		*pv = NULL;

	/* look up id */
	p.ptr = htsp_get(procs, (char *)id);
	if (p.ptr == NULL)
		return -1;

	/* load output pointers */
	if ((p.v->h.virt) && (pv != NULL))
		*pv = p.v;
	else if (pr != NULL)
		*pr = p.r;

	/* return type */
	return p.v->h.virt;
}

proc_virt_t *proc_lookup_safe(const char *id)
{
	proc_virt_t *p;
	p = htsp_get(procs, (char *)id);
	if (p == NULL)
		plumb_error("Process %s does not exist", id);
	return p;
}

/* ==== administrative ==== */

void proc_start(proc_virt_t *proc)
{
	proc_virt_t *pv;
	proc_real_t *pr;

	pv = proc;
	pr = (proc_real_t *)proc;
	if (pv->h.state == PS_STOPPED) {
		if (pv->h.virt) {
			if (pv->start != NULL)
				vp_act(pv, pv->start(pv));
			pv->h.state = PS_RUNNING;
		}
		else
			rp_start(pr);
	}
	if (pv->h.sticky)
		plumb_sticky_started();
}

void proc_stop(proc_virt_t *proc, int hard)
{
	if (proc->h.state == PS_STOPPED)
		return;
	if (proc->h.virt)
		vp_stop(proc, hard);
	else
		rp_stop((proc_real_t *)proc, hard);
}

void proc_pause(proc_virt_t *proc, int pause)
{
	int n;

	if (proc->h.state == PS_STOPPED)
		return;

	if (pause && (proc->h.state == PS_PAUSED))
		return;
	if (!pause && (proc->h.state != PS_PAUSED))
		return;

	for(n = 0; n < proc->h.num_fds; n++)
		fd_pause(&(proc->h.fds[n]), pause);

	if (pause)
		proc->h.state = PS_PAUSED;
	else
		proc->h.state = PS_RUNNING;
}


void proc_start_all(void)
{
	htsp_entry_t *e;

	for (e = htsp_first(procs); e; e = htsp_next(procs, e))
		proc_start(e->value);
}

void proc_stop_all(int hard, int virt)
{
	htsp_entry_t *e;
	proc_virt_t *p;

	for (e = htsp_first(procs); e; e = htsp_next(procs, e)) {
		p = e->value;
		if ((virt) || (!p->h.virt))
			proc_stop(e->value, hard);
	}
}


void proc_free_all(void)
{
	htsp_entry_t *e;

	for (e = htsp_first(procs); e; e = htsp_next(procs, e))
		proc_free(e->value);

	htsp_free(procs);
	procs = NULL;
}

void proc_close_stdio(proc_virt_t *p)
{
	int n;
	for(n = 0; (n < 3) && (n < p->h.num_fds); n++)
		fd_close(&p->h.fds[n]);
}

void proc_close_stdio_all(int close_virt)
{
	htsp_entry_t *e;
	proc_virt_t *p;

	for (e = htsp_first(procs); e; e = htsp_next(procs, e)) {
		p = e->value;
		if ((!p->h.virt) || close_virt)
			proc_close_stdio(e->value);
	}
}

int proc_count_real_running(void)
{
	htsp_entry_t *e;
	proc_virt_t *p;
	int cnt;

	cnt = 0;
	for (e = htsp_first(procs); e; e = htsp_next(procs, e)) {
		p = e->value;
		if ((!p->h.virt) && (p->h.state != PS_STOPPED))
			cnt++;
	}
	return cnt;
}


/* ====== process properties ====== */
typedef enum {
	PT_INVALID = 0,
	PT_BOOL
} prop_type_t;

struct prop_s {
	char *name;
	prop_type_t type;
	int offs;
};

/* we need an existing process to get field pointers in a non-gcc-only way;
   this one happens to exist already. */
extern proc_virt_t *procwatch;

static struct prop_s proc_props[] = {
	/* name         type           offs */
	{ "sticky",     PT_BOOL,       offsetof(proc_header_t, sticky)},
	{ "kill_on_eof",PT_BOOL,       offsetof(proc_header_t, kill_on_eof)},
	{ NULL,         PT_INVALID,    0}
};

int bool2int(const char *str)
{
	if ((strcmp(str, "yes") == 0) || (strcmp(str, "true") == 0) ||
	    (strcmp(str, "on") == 0)  || (strcmp(str, "1") == 0))
	    return 1;
	if ((strcmp(str, "no") == 0)  || (strcmp(str, "false") == 0) ||
	    (strcmp(str, "off") == 0) || (strcmp(str, "0") == 0))
	    return 0;
	plumb_error("Invalid bool value %s, should be true or false\n", str);
	return -1; /* will never return since plumb_error() aborts */
}

static void proc_prop_set(proc_virt_t *proc, kvlist_t *props)
{
	kv_t *prop;
	struct prop_s *p;
	char *header;

	if (props == NULL)
		return;

	header = (char *)&proc->h;
	for(prop = props->head; prop != NULL; prop = prop->next) {
	/* for this small amount of properties linear search is ok;
	   might need to switch to an sphash later */
		for(p = proc_props; p->name != NULL; p++)
			if (strcmp(p->name, prop->key) == 0)
				goto found;
		plumb_error("Unknown process property: %s\n", prop->key);
		found:;
		switch(p->type) {
			case PT_BOOL:
				*(char *)(header + p->offs) = bool2int(prop->value);
				break;
			case PT_INVALID:
				plumb_error("Internal error: invalid process property type\n");
				abort();
		}
	}
}


/* has to be here instead of fd.c because of the process lookup */
fd_t *fd_ref(char *proc_id, int proc_fd)
{
	proc_t p;

	procs_init();

	p.ptr = htsp_get(procs, (char *)proc_id);
	if (p.ptr == NULL)
		plumb_error("can not find process %s\n", proc_id);

	return fd_ref_(p.v, proc_fd);
}

void proc_kill_on_eof(proc_virt_t *p)
{
	int active_sinks, n;

	if (!p->h.kill_on_eof)
		return;

	/* count active (non-eof'd) sinks */
	active_sinks = 0;
	for(n = 0; n < p->h.num_fds; n++) {
		fd_t *fd = &p->h.fds[n];
		if (is_sink(fd) && !fd->eof)
			active_sinks++;
	}

	if (active_sinks == 0)
		proc_stop(p, 0);
#warning TODO: wait some timeout and do a hard kill
}

