/*  plumb - the pipe shell - split stream by a set of characters
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/
#include <stdlib.h>
#include <string.h>
#include "process.h"
#include "fd.h"
#include "libporty/util/buf.h"
#include "virt.h"
#include "config.h"

typedef struct msg_s msg_t;

struct msg_s {
	char *data;
	int len;
	msg_t *next;
};

typedef struct {
	char *err_msg_len_overrun;
	char *err_msg_num_overrun;
	int totlen, maxlen;
	int totnum, maxnum;
	msg_t *head, *tail;  /* append incoming messages at the end at tail, pop messages from head */
	char event;
} tank_t;

static void append(tank_t *t, msg_t *m)
{
	if (t->tail != NULL)
		t->tail->next = m;
	t->tail = m;
	if (t->head == NULL)
		t->head = m;
}

static virt_ctrl_t tank_input(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	tank_t *t = proc->data;
	msg_t *m;

	/* simplest case: empty buffer, no block - just write */
	if ((!vp_is_blocked(proc, vstdout)) && (t->head == NULL)) {
#ifdef TANK_DEBUG
		fprintf(stderr, "tank %s {%d/%d}: flow through %d\n", proc->h.id, t->totnum, t->totlen, len);
#endif
		vp_write(proc, vstdout, data, len);
		return VC_OK;
	}

	if (t->totnum+1 >= t->maxnum) {
		if (t->event) {
			char *err;
			err = malloc(strlen(proc->h.id) + 64);
			len = sprintf(err, "plumb_tank %s buffer has too many datagrams\n", proc->h.id);
			vp_write(proc, vstdevent, err, len);
			free(err);
		}
		if (t->err_msg_num_overrun != NULL)
			vp_write(proc, vstderr, t->err_msg_num_overrun, strlen(t->err_msg_num_overrun));
		return VC_QUIT;
	}

	if (t->totlen+len >= t->maxlen) {
		if (t->event) {
			char *err;
			err = malloc(strlen(proc->h.id) + 64);
			len = sprintf(err, "plumb_tank %s buffer grew too large\n", proc->h.id);
			vp_write(proc, vstdevent, err, len);
			free(err);
		}
		if (t->err_msg_len_overrun != NULL)
			vp_write(proc, vstderr, t->err_msg_len_overrun, strlen(t->err_msg_num_overrun));
		return VC_QUIT;
	}

	/* else we need to append to the buffer chain */
	m = malloc(sizeof(msg_t));
	m->data = malloc(len);
	memcpy(m->data, data, len);
	m->len = len;
	m->next = NULL;
	append(t, m);

	t->totlen += len;
	t->totnum++;

#ifdef TANK_DEBUG
		fprintf(stderr, "tank %s {%d/%d}: appended %d\n", proc->h.id, t->totnum, t->totlen, m->len);
#endif

	return VC_OK;
}

static void tank_free(proc_virt_t *proc)
{
	tank_t *t = proc->data;
	msg_t *m, *next;

	for(m = t->head; m != NULL; m = next) {
		next = m->next;
		if (m->data != NULL)
			free(m->data);
		free(m);
	}

	if (t->err_msg_num_overrun != NULL)
		free(t->err_msg_num_overrun);
	if (t->err_msg_len_overrun != NULL)
		free(t->err_msg_len_overrun);
	free(t);
}

virt_ctrl_t tank_source_block(proc_virt_t *proc, int local_fd, int is_blocking)
{
	msg_t *next;
	tank_t *t = proc->data;
	int got_eof = 0;

	if (is_blocking)
		return VC_OK;

	for(; t->head != NULL; t->head = next) {
		next = t->head->next;

#ifdef TANK_DEBUG
		fprintf(stderr, "tank %s {%d/%d}: remove %d\n", proc->h.id, t->totnum, t->totlen, t->head->len);
#endif
		if (t->head->data != NULL) {
			vp_write(proc, vstdout, t->head->data, t->head->len);
			t->totlen -= t->head->len;
			t->totnum--;
			free(t->head->data);
		}
		else
			got_eof = 1;
		free(t->head);

		if (got_eof) {
			t->head = next;
			return VC_QUIT;
		}
		if (vp_is_blocked(proc, vstdout)) {
#ifdef TANK_DEBUG
			fprintf(stderr, "tank %s {%d/%d} got blocked again\n", proc->h.id, t->totnum, t->totlen);
#endif
			break;
		}
	}

	if (t->head == NULL) {
		t->tail = NULL;
#ifdef TANK_DEBUG
		fprintf(stderr, "tank %s {%d/%d}: empty\n", proc->h.id, t->totnum, t->totlen);
#endif
	}

	return VC_OK;
}

/* eof on sink (input): may need to buffer it */
virt_ctrl_t tank_sink_eof(proc_virt_t *proc, int local_fd)
{
	tank_t *t = proc->data;
	msg_t *m;

	if (t->head == NULL) /* no pending data in the buffer, just quit */
		return VC_QUIT;

	/* pending output - append an empty packet which will serve as an EOF */
	m = calloc(sizeof(msg_t), 1);
	append(t, m);
	return VC_OK;
}

int vp_tank_init(proc_virt_t *p, kvlist_t *args)
{
	tank_t *s;
	char *len_overrun = NULL, *num_overrun = NULL;
	char *smaxlen = NULL, *smaxnum = NULL;
	char *sevent = NULL;
	vp_args_t vargs[] = {
		{"len_overrun_msg", &len_overrun},
		{"num_overrun_msg", &num_overrun},
		{"maxlen",          &smaxlen},
		{"maxnum",          &smaxnum},
		{"event",           &sevent},
		{NULL, NULL}
	};

	vp_args(args, vargs);

	s = calloc(1, sizeof(tank_t));
	s->event = vp_istrue(sevent);
	s->err_msg_len_overrun  = strclone(len_overrun);
	s->err_msg_num_overrun  = strclone(num_overrun);

	if (smaxlen != NULL)
		s->maxlen = vp_arg_atoi("[tank]", "maxlen", smaxlen);
	else
		s->maxlen = VP_TANK_DEFAULT_MAX_LEN;

	if (smaxnum != NULL)
		s->maxnum = vp_arg_atoi("[tank]", "maxnum", smaxnum);
	else
		s->maxnum = VP_TANK_DEFAULT_MAX_NUM;

	p->input        = tank_input;
	p->free         = tank_free;
	p->source_block = tank_source_block;
	p->sink_eof     = tank_sink_eof;
	p->data         = s;

	vp_stdfds_set(p, VP_REQ_QUIT_ON_UNBIND, VP_REQ_QUIT_ON_UNBIND, VP_REQ_UNBOUND, VP_REQ_UNBOUND, s->event ? VP_REQ_QUIT_ON_UNBIND : VP_REQ_UNBOUND, VP_REQ_UNBOUND);

	return 0;
}
