/*  plumb - the pipe shell - split stream into variable length binary records using header information
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/
#include <stdlib.h>
#include <string.h>
#include "process.h"
#include "fd.h"
#include "virt.h"
#include "config.h"

typedef struct {
	/* config */
	int field_offs;
	char field_len;
	char LSB;
	int len_delta;
	int maxlen;

	/* states */
	int len;              /* full len of the current packet or -1 if not yet known */
	int used, alloced;
	char *buff;
} varlen_t;

/* append len bytes of data to buffer of a; if datalen is not NULL, modify data
   and datalen by len; len must be less than datalen */
static void append(varlen_t *a, const char **data, int *datalen, int len)
{
	if (a->used + len >= a->alloced) {
		a->alloced += len + 256;
		a->buff = realloc(a->buff, a->alloced);
	}
	memcpy(a->buff + a->used, *data, len);
	a->used += len;

	*data += len;
	*datalen -= len;
}

/* design:
   Buffer only one full record. Input data in 2 phases:
    1. read input until the length field is known;
    2. read input exactly until the full record is loaded.
   Both phases should handle short reads.
   After sending out a record, the input buffer may contain (first portion of)
   the next record - restart reading the same way.
*/
static virt_ctrl_t varlen_input(proc_virt_t *proc, int local_fd, const char *data, int len)
{
	varlen_t *a = proc->data;
	int n, need, cpy, dir;
	unsigned char *c;

	do {
		/* do we already know the length of the record? */
		if (a->len < 0) {
			/* no, check if we can read just enough to have that */
			need = a->field_offs + a->field_len - a->used;
			cpy = len < need ? len : need;
			append(a, &data, &len, cpy);

			/* if we don't have enough data to determine len, exit now */
			if (cpy < need)
				return VC_OK;
		
			/* set length */
			a->len = 0;
			if (a->LSB) {
				c = (unsigned char *)(a->buff) + a->field_offs + a->field_len;
				dir = -1;
			}
			else {
				c = (unsigned char *)(a->buff) + a->field_offs;
				dir = 1;
			}
			for(n = 0; n < a->field_len; n += dir, c += dir) {
				a->len <<= 8;
				a->len |= *c;
			}

			a->len += a->len_delta;
			if (a->len < 0) {
				vp_printf(proc, vstderr, 128, "invalid length: negative (%d)\n", a->len);
				return VC_QUIT;
			}
			if (a->len > a->maxlen) {
				vp_printf(proc, vstderr, 128, "invalid length: too big (%d > %d)\n", a->len, a->maxlen);
				return VC_QUIT;
			}
		}

		/* at this point we have a partial record with a known length
		   grow it until it is a full record, then restart */
		need = a->len - a->used;
		cpy = len < need ? len : need;
		append(a, &data, &len, cpy);

		/* if we don't have enough data to finish the record, return to wait for more */
		if (cpy < need)
			return VC_OK;

		/* we've just finished reading a record: send it out and reset */
		vp_write(proc, 1, a->buff, a->len);
		a->used = 0;
		a->len = -1;

		/* and start the whole thing again, if we still have something in the input buffer */
	} while (len > 0);
	return VC_OK;
}

void varlen_free(proc_virt_t *p)
{
	varlen_t *a = p->data;
	if (a->buff != NULL)
		free(a->buff);
	free(a);
}

int vp_varlen_init(proc_virt_t *p, kvlist_t *args)
{
	varlen_t *a;
	int len, maxlen, offs, len_delta, lsb;
	char *slen = NULL, *slen_delta = NULL, *smaxlen = NULL, *soffs = NULL, *sorder = NULL;
	vp_args_t vargs[] = {
		{"offs",      &soffs},
		{"size",      &slen},
		{"len_delta", &slen_delta},
		{"order",     &sorder},
		{"maxlen",    &smaxlen},
		{NULL, NULL}
	};

	vp_stdfds_set(p, VP_REQ_QUIT_ON_UNBIND, VP_REQ_QUIT_ON_UNBIND, VP_REQ_ANY, VP_REQ_UNBOUND, VP_REQ_UNBOUND, VP_REQ_UNBOUND);

	vp_args(args, vargs);

	if (slen == NULL)
		plumb_error("varlen %s without size\n", p->h.id);

	len  = vp_arg_atoi("[varlen]", "size", slen);
	offs = vp_arg_atoi("[varlen]", "offs", soffs);
	
	if (slen_delta != NULL)
		len_delta = vp_arg_atoi("[varlen]", "slen_delta", slen_delta);
	else
		len_delta = 0;

	if ((len < 1) || (len > 4))
		plumb_error("varlen %s: len out of range (must be between 1 and 4)\n", p->h.id);

	if (smaxlen != NULL)
		maxlen = vp_arg_atoi("[varlen]", "maxlen", smaxlen);
	else
		maxlen = VP_VARLEN_DEFAULT_MAX_LEN;

	lsb = 0;
	if (sorder != NULL) {
		if ((*sorder == 'L') || (*sorder == 'l'))
			lsb =1;
		else if ((*sorder == 'M') || (*sorder == 'm'))
			lsb = 0;
		else
			plumb_error("varlen %s: order must be LSB or MSB", p->h.id);
	}

	if (offs + len >= maxlen)
		plumb_error("[varlen] maxlen is too small: length field would not fit\n");

	a = calloc(1, sizeof(varlen_t));
	a->field_len  = len;
	a->field_offs = offs;
	a->len_delta  = len_delta;
	a->LSB        = lsb;
	a->maxlen     = maxlen;
	a->len        = -1; /* length of the current record is not known */

	p->input = varlen_input;
	p->free  = varlen_free;
	p->data  = a;

	return 0;
}
