/*  plumb - the pipe shell - heleprs for virtual processes (VPs)
    Copyright (C) 2012-2013  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Project page: http://repo.hu/projects/plumb
    Contact: plumb (at) igor2.repo.hu
*/
#include <stdarg.h>
#include "virt.h"
#include "pipe.h"

void vp_args(kvlist_t *args, vp_args_t *desc)
{
	kv_t *kv;
	vp_args_t *d;

	if (args == NULL)
		return;

	/* typically we have only a couple of arguments and descriptions and keys
	   differ in the first few chars and vp_args() is ran only on init(), so
	   the following slow algorithm is good enough. */

	for(kv = args->head; kv != NULL; kv = kv->next)
		for(d = desc; d->name != NULL; d++)
			if (strcmp(d->name, kv->key) == 0)
				*d->value = kv->value;
}

int vp_arg_atoi(const char *vpname, const char *name, const char *val)
{
	int i;
	char *end;
	if (val == NULL)
		plumb_error("Argument %s should not be empty for %s\n", name, vpname);

	i = strtol(val, &end, 10);
	if (*end != '\0')
		plumb_error("Argument %s should be an floating point number for %s\n", name, vpname);

	return i;
}

double vp_arg_atof(const char *vpname, const char *name, const char *val)
{
	double d;
	char *end;
	if (val == NULL)
		plumb_error("Argument %s should not be empty for %s\n", name, vpname);

	d = strtod(val, &end);
	if (*end != '\0')
		plumb_error("Argument %s should be an integer for %s\n", name, vpname);

	return d;
}


void vp_stdfds_set(proc_virt_t *proc, vp_req_t stdin_, vp_req_t stdout_, vp_req_t stderr_, vp_req_t stdctrl, vp_req_t stdevent, vp_req_t rest)
{
	proc->stdfds[0] = stdin_;
	proc->stdfds[1] = stdout_;
	proc->stdfds[2] = stderr_;
	proc->stdfds[3] = stdctrl;
	proc->stdfds[4] = stdevent;
	proc->stdfds[5] = rest;

	/* do default binding to stdevent if it shall be kept open and not bound already */
	if (vp_stdfd_want(proc, vstdevent) && !(vp_is_bound(proc, vstdevent)))
		vp_pipe_stdevent(proc)
}

virt_ctrl_t vp_stdfd_enforce(proc_virt_t *proc, int local_fd, int open)
{
	int idx = local_fd;
	
	if (idx > vstdevent + 1)
		idx = vstdevent + 1;
	switch(proc->stdfds[idx]) {
		case VP_REQ_ANY:
			break;
		case VP_REQ_UNBOUND:
			if (open)
				plumb_error("Virtual process %s ([%s]) can not have fd %d bound", proc->h.id, proc->cmd, local_fd);
			break;
		case VP_REQ_ERR_ON_UNBIND:
			if (!open)
				plumb_error("Virtual process %s ([%s]) requires fd %d to be bound", proc->h.id, proc->cmd, local_fd);
			break;
		case VP_REQ_QUIT_ON_UNBIND:
			if (!open)
				return VC_QUIT;
	}
	return VC_OK;
}

void vp_shadow_rebind(proc_virt_t *vp, int vp_fd, proc_real_t *shadow, int real_fd)
{
	fd_t *remote;
	int src;

	/* allocate shadow fd */
	fd_ref_((proc_virt_t *)shadow, vp_fd);
	shadow->sock[vp_fd] = real_fd;
	fd_set_socket(&shadow->h.fds[vp_fd], real_fd);

	/* determine remote end of the pipe and recalculate POLLIN for local end if needed */
	if (vp_is_source(vp, vp_fd)) {
		fd_rp_source_poll(&shadow->h.fds[vp_fd]);
		remote = vp->h.fds[vp_fd].pipe->to;
		src = 1;
	}
	else {
		remote = vp->h.fds[vp_fd].pipe->from;
		src = 0;
	}

	/* we don't have to kill and rebuild the pipe as only one end changes -
	   if the other (static) end is a real process, it won't feel the difference
	   and won't panic */
	if (src)
		pipe_rebind(vp->h.fds[vp_fd].pipe, &shadow->h.fds[vp_fd], remote);
	else
		pipe_rebind(vp->h.fds[vp_fd].pipe, remote, &shadow->h.fds[vp_fd]);
}

proc_real_t *vp_shadow_create(proc_virt_t *p, char *name)
{
	proc_real_t *shadow;

	if (name == NULL) {
		name = malloc(strlen(p->cmd)+strlen(p->h.id)+16);
		sprintf(name, "%s_%s_SHADOW_%d", p->h.id, p->cmd, p->num_shadows++);
	}
	shadow = (proc_real_t *)proc_create(p->tx, name, 0, NULL, NULL, NULL);
	shadow->h.state = PS_RUNNING;    /* pretend the process is already running so pr_start won't try to start it */
	shadow->h.sticky = p->h.sticky;  /* inherit stickyness from shadowed process */
	shadow->accept_fd = 1;           /* we will do a lot of rebinding */
	shadow->shadow_of = p;
	return shadow;
}

void vp_shadow_create_finish(proc_real_t *shadow, int accept_fd)
{
	shadow->accept_fd = accept_fd;
}

void vp_printf(proc_virt_t *proc, int fd, int maxlen, char *fmt, ...)
{
	va_list ap;
	int len;
	char *buff;

	buff = malloc(maxlen);
	va_start(ap, fmt);
	len = vsprintf(buff, fmt, ap);
	va_end(ap);

	vp_write(proc, fd, buff, len);
	free(buff);
}

