#include "process.h"
#include "fd.h"
#include "kv.h"
#include "plumb.h"

/* Write data to an integer fd, if that fd exist */
#define vp_write(proc, fd, data, len) \
	do { \
		if (proc->h.num_fds > (fd)) \
			fd_write(&(proc->h.fds[fd]), (data), (len)); \
	} while(0)

/* close an fd if not already closed */
#define vp_close(proc, fd) \
	do { \
		if (proc->h.num_fds > (fd)) \
			fd_close(&(proc->h.fds[fd])); \
	} while(0)


enum vp_default_fds {
	vstdin    = 0,
	vstdout   = 1,
	vstderr   = 2, /* printing non-critical runtime errors */
	vstdctrl  = 3, /* reading control commands */
	vstdevent = 4, /* emitting events */
};

/* connect stdevent output of a process to the event handler to allow
   the process to emit events */
#define vp_pipe_stdevent(p) \
	pipe_register(fd_ref_(p, vstdevent), fd_ref("event", -1));


/* true if fd is blocked or not bound */
#define vp_is_blocked(proc, fd) (((proc)->h.num_fds <= (fd)) || ((proc)->h.fds[fd].pipe == NULL) || (proc->h.fds[fd].blocked) || (proc->h.fds[fd].paused))

/* true if fd is bound */
#define vp_is_bound(proc, fd) (((proc)->h.num_fds > (fd)) && ((proc)->h.fds[fd].pipe != NULL))

/* true if fd is a sink or source for proc */
#define vp_is_sink(proc, fd) (((proc)->h.num_fds > (fd)) && (is_sink(&(proc)->h.fds[fd])))
#define vp_is_source(proc, fd) (((proc)->h.num_fds > (fd)) && (is_source(&(proc)->h.fds[fd])))


/* return true if s is not NULL and starts with [yYtT1] */
#define vp_istrue(s) (((s) != NULL) && ((*(s) == 'y') || (*(s) == 'Y') || (*(s) == 't') || (*(s) == 'T') || (*(s) == '1')))

/* ==== standard argument parsing ==== */

typedef struct vp_args_s {
	char *name;
	char **value;
} vp_args_t;


void vp_args(kvlist_t *args, vp_args_t *desc);

int vp_arg_atoi(const char *vpname, const char *name, const char *val);
double vp_arg_atof(const char *vpname, const char *name, const char *val);


/* ==== standard fd binding functions ==== */

typedef enum {
	VP_REQ_ANY = 0,          /* do not care (might be used for non-standard things) */
	VP_REQ_UNBOUND,          /* error if fd gets bound (piped) */
	VP_REQ_ERR_ON_UNBIND,    /* error if unbound */
	VP_REQ_QUIT_ON_UNBIND    /* quit if unbound */
} vp_req_t;

/* set behavior of default fds - will be used by default callbacks */
void vp_stdfds_set(proc_virt_t *proc, vp_req_t stdin_, vp_req_t stdout_, vp_req_t stderr_, vp_req_t stdctrl, vp_req_t stdevent, vp_req_t rest);

/* retruns VC_OK/VC_QUIT or generates error if local_fd for proc is open or
   closed in a way proc->stdfds does not permit. Argument open should be 1
   for newfd (or when the fd is open in general) and 0 for delfd (or when
   the fd is closed). */
virt_ctrl_t vp_stdfd_enforce(proc_virt_t *proc, int local_fd, int open);

/* true if we want p:fd open explicitly */
#define vp_stdfd_want(p, fd) (((p)->stdfds[fd] == VP_REQ_ERR_ON_UNBIND) || ((p)->stdfds[fd] == VP_REQ_QUIT_ON_UNBIND))


/* === shadow functions === */

/* start creating a new shadow process for p; if name is NULL, generate an unique name, else
   name must be strclone()'d */
proc_real_t *vp_shadow_create(proc_virt_t *p, char *name);

/* rebind vp:vp_fd to a shadow RP and put a real fd behind it */
void vp_shadow_rebind(proc_virt_t *vp, int vp_fd, proc_real_t *shadow, int real_fd);

/* finish creating a new shadow process; accept_fd should be 1 if
   vp_shadow_rebind() will be called to this shadow in the future */
void vp_shadow_create_finish(proc_real_t *shadow, int accept_fd);


/* printf to an fd - SLOW */
void vp_printf(proc_virt_t *proc, int fd, int maxlen, char *fmt, ...);

